#
# shinGETsu - P2P anonymous BBS
#
# (c) Fuktommy <fuktommy@users.sourceforge.net>
# Released under the GNU General Public License
# $Id: NodeList.pm,v 1.8 2004/08/24 16:06:45 fuktommy Exp $
#

package Shingetsu::NodeList;
use Socket;
use Shingetsu::Config;
use Shingetsu::Node;
use Shingetsu::Util;

#
# configration
#
$nodelist = $Shingetsu::Config::nodelist;

#
# show all nodes
#
sub list() {
	local $_;
	open IN, $nodelist or return ();
	my @node = ();
	while (<IN>) {
		chomp;
		my $n = new Shingetsu::Node($_);
		push @node, $n;
	}
	return @node;
}

#
# show randmized all nodes
#
sub all() {
	local $_;
	my @node = list();
	for ($i=@node; $i>0; $i--) {
		my $t = int rand $i;
		($node[$i-1], $node[$t]) = ($node[$t], $node[$i-1]);
	}
	return @node;
}

#
# suggest random node
#
sub random() {
	local $_;
	my @node = list();
	if (@node) {
		my $node = $node[rand(0+@node)];
		return $node;
	} else {
		return undef;
	}
}

#
# who am i
#
sub myself() {
	my $port = $ENV{SERVER_PORT};
	local $_;
	my @node = all();
	my $path = $Shingetsu::Config::path;
	if ($ENV{SERVER_NAME}) {
		return new Shingetsu::Node($ENV{SERVER_NAME}, $port, $path);
	}
	foreach (@node) {
		my @buf = $_->talk("/ping");
		if ((defined $buf[1])&&($buf[0] eq "PONG")) {
			return new Shingetsu::Node($buf[1], $port, $path);
		}
	}
	return undef;
}

#
# add node to list
# args: node
#
sub add($) {
	my($node) = @_;
	my $str = $node->toString() . "\n";
	local $_;
	unless (include($node)) {
		Shingetsu::Util::lock();
		open OUT, ">> $nodelist" or die "$nodelist: open failed. $!";
		print OUT "$str";
		close OUT;
		Shingetsu::Util::unlock();
	}
}

#
# include node
# args: node
#
sub include($) {
	my($node) = @_;
	my $str = $node->toString();
	local $_;
	my @node = list();
	if (grep $_->toString() eq $str, @node) {
		return 1;
	} else {
		return undef;
	}
}

#
# count node
#
sub count() {
	my @node = list();
	return 0+@node;
}

#
# remove node from list
# args: node
#
sub remove($) {
	my($node) = @_;
	my $str = $node->toString();
	$str .= "\n";
	local $_;
	my @node;
	if (-e $nodelist) {
		open IN, $nodelist or exit;
		@node = <IN>;
		close IN;
	} else {
		return;
	}
	if (grep $_ eq "$str", @node) {
		Shingetsu::Util::lock();
		open OUT, "> $nodelist" or die "$nodelist: open failed. $!";
		foreach (@node) {
			print OUT if ($_ ne $str);
		}
		close OUT;
		Shingetsu::Util::unlock();
	}
}

#
# ping all nodes
#
sub pingall() {
	local $_;
	my @node = all();
	foreach (@node) {
		unless ($_->ping()) {
			remove($_);
		}
	}
}

#
# join network
# args: node, port
#
sub join($) {
	my ($node) = @_;
	my $port = $ENV{SERVER_PORT};
	my $count = 0;
	while (1) {
		my @buf = $node->join();
		my $n = $node->toString();
		if ((@buf == 1)&&($buf[0] eq "WELCOME")) {
			add($node);
			last;
		} elsif ($count++ > $Shingetsu::Config::retrySearch) {
			last;
		} elsif ((@buf == 2)&&($buf[0] eq "WELCOME")) {
			add($node);
			$node = new Shingetsu::Node($buf[1]);
		} else {
			last;
		}
	}
}

#
# connect initial node
# args: port
#
sub init() {
	my $port = $ENV{SERVER_PORT};
	my @initNode = @Shingetsu::Config::initNode;
	unshift @initNode, split(/\+/, $ENV{X_INIT_NODE}) if ($ENV{X_INIT_NODE});
	local $_;
	my $inode;
	foreach (@initNode) {
		$inode = new Shingetsu::Node($_);
		if ($inode->ping()) {
			Shingetsu::NodeList::join($inode);
			last;
		}
	}
	my $myself = myself();
	remove($myself) if ($myself);
	my @node = all();
	my %done = ();
	return if (@node == 0);

	$done{$_->toString()} = 1 foreach (@node);
	while (1) {
		my $node = $node[rand(0+@node)];
		my @buf  = $node->talk("/node");
		my $str = "";
		if ((@buf > 0)&&($buf[0] =~ m|:\d+/[^: ]+$|)) {
			$str = $buf[0];
			$node = new Shingetsu::Node($str);
			push @node, $node unless ($done{$str});
		}
		if (($done{$str}++ > $Shingetsu::Config::retrySearch)||
				(@node >= $Shingetsu::Config::nodes)) {
			last;
		}
	}

	foreach (@node) {
		Shingetsu::NodeList::join($_) unless (include($_));
	}
	remove($myself) if ($myself);
	if (count() >= $Shingetsu::Config::nodes) {
		$inode->bye();
		remove($inode);
	}
	while (count() > $Shingetsu::Config::nodes) {
		my $node = random();
		$node->bye();
		remove($node);
	}
	warn "NodeList::init() finished\n";
}

1;
